### CRUNCH - debugger UI

set tag_stack {}
set font_name Courier
set font_size 7
# nid-> tid
set nodedict ""
# tid -> typename
set typedict ""
# nid -> fn:ln
set locdict ""
# fn:ln -> {nid ...}
set reverse_locdict ""
# nid -> {nid ...}
set inner_nodes ""
set current_filename ""
# fn -> text
set pseudo_files ""
# id -> data
set lambdadict ""
set all_files {}

proc setup_ui {} {
	global font_name font_size
	frame .left
	pack .left -side left -fill both -expand 1
	frame .right
	pack .right -side right -fill both -expand 1
	label .left.type -text "\n" -background white -font [list $font_name $font_size]
	pack .left.type -side top -fill x
	text .left.nodes -wrap char -font [list $font_name $font_size] \
		-yscrollcommand {.left.nscroll set}
	scrollbar .left.nscroll -command {.left.nodes yview}
	pack .left.nscroll -fill y -side right
	pack .left.nodes -side top -fill both -expand 1
	ttk::combobox .right.files -font [list $font_name $font_size]
	pack .right.files -side top -fill x
	panedwindow .right.box -orient vertical
	pack .right.box -side top -fill both -expand 1
	frame .right.box.source
	pack .right.box.source -side top -fill both -expand 1
	scrollbar .right.box.source.scroll -command {.right.box.source.code yview}
	pack .right.box.source.scroll -fill y -side right
	text .right.box.source.code -wrap char -font [list $font_name $font_size] \
		-yscrollcommand {.right.box.source.scroll set}
	.right.box.source.code tag configure location -background lightgoldenrod
	pack .right.box.source.code -side left -fill both -expand 1
	frame .right.box.events
	pack .right.box.events -side top -fill both -expand 1
	scrollbar .right.box.events.scroll -command {.right.box.events.text yview}
	pack .right.box.events.scroll -fill y -side right
	text .right.box.events.text -wrap char -font [list $font_name $font_size] \
		-yscrollcommand {.right.box.events.scroll set}
	.right.box.events.text tag configure title -underline 1
	.right.box.events.text tag configure error -background mistyrose1
	pack .right.box.events.text -side top -fill both -expand 1
	.right.box add .right.box.source .right.box.events
	bind . <Escape> exit
	bind .right.files <<ComboboxSelected>> select_file
	bind .right.box.source.code <Motion> {select_loc %x %y}
	bind .right.box.source <Leave> unhighlight
	bind .left.nodes <ButtonPress-3> {search_next %W %x %y}
	bind .right.box.source.code <ButtonPress-3> {search_next %W %x %y}
	bind .right.box.events.text <ButtonPress-3> {search_next %W %x %y}
}

proc search_next {w x y} {
	if {![catch {$w index "@$x,$y"} p1]} {
		set p2 [$w search -backwards -regexp {^|[][("',;#){}]|[[:space:]]} $p1]
		set p3 [$w search -regexp {^|[][("',;#){}]|[[:space:]]} "$p1+1 chars"]
		set txt [$w get "$p2+1 chars" $p3]
		puts stderr ":$txt:"
		set p [$w search -exact -count n $txt $p3]
		if {$p != ""} {
			$w tag remove sel 1.0 end
			$w tag add sel $p "$p+$n chars"
			lassign [$w dlineinfo $p] xp yp
			$w see $p
			warp_to_index $w $p
		}
	}
}

proc warp_to_index {w index} {
    set info [$w bbox $index]
    if {$info != ""} {
        set x [expr [lindex $info 0] + [lindex $info 2] / 2]
        set y [expr [lindex $info 1] + [lindex $info 3] / 2]
        event generate $w <Motion> -x $x -y $y -warp 1
    }
}

proc select_node {nid} {
	global typedict nodedict
	.left.nodes tag configure $nid -background darkolivegreen3
	set tid [dict get $nodedict $nid]
	.left.type configure -text "$tid: [dict get $typedict $tid]" -background darkolivegreen3
}

proc mark_src {loc loadfile} {
	global current_filename
	if {[regexp {^([^:]+):(\d+)$} $loc _ fname line]} {
		if {$current_filename == $fname || ($loadfile && [ensure_file $fname])} {
			.right.box.source.code tag add location $line.0 "$line.0 lineend+1 chars"
		}
	}
}

proc node_enter {nid} {
	global typedict inner_nodes locdict nodedict
	select_node $nid
	if {[dict exists $inner_nodes $nid]} {
		foreach n [dict get $inner_nodes $nid] {
			set tid [dict get $nodedict $n]
			if {[dict get $typedict $tid] != "?"} {
				.left.nodes tag configure $n -background darkolivegreen1
			}
		}
	}
	if {[dict exists $locdict $nid]} {
		set loc [dict get $locdict $nid]
		.right.box.source.code tag remove location 1.0 end
		mark_src $loc 1
	}
}

proc node_leave {nid} {
	global typedict inner_nodes nodedict
	default_color $nid
	if {[dict exists $inner_nodes $nid]} {
		foreach n [dict get $inner_nodes $nid] {
			default_color $n
		}
	}
	.left.type configure -text "\n" -background white
}

proc eval_input {} {
	if {[gets stdin input] >= 0} {
		eval $input
	} elseif {[eof stdin]} {
		fileevent stdin readable ""
	}
}

proc debugging {} {
	global typedict locdict tag_stack inner_nodes all_files current_filename
	global lambdadict reverse_locdict nodedict
	.right.box.events.text delete 1.0 end
	.left.nodes delete 1.0 end
	set typedict [dict create]
	set nodedict [dict create]
	set locdict [dict create]
	set reverse_locdict [dict create]
	set tag_stack {}
	set inner_nodes [dict create]
	set lambdadict [dict create]
	set current_filename ""
	set all_files {}
}

proc msg {str} {
	set tag {}
	if {[string match {Error:*} $str]} {set tag error}
	.right.box.events.text insert end $str $tag
	.right.box.events.text see end
}

proc txt {str} {
	global tag_stack
	set tag {}
	if {$tag_stack != ""} {set tag [lindex $tag_stack end]}
	.left.nodes insert end $str $tag
}

proc default_color {nid} {
	global typedict nodedict
	set col white
	set tid [dict get $nodedict $nid]
	if {[dict get $typedict $tid] == "?\n"} {
		set col indianred1
	}
	.left.nodes tag configure $nid -background $col
}

proc btags {nid tid type vs {loc ""}} {
	global tag_stack nodedict typedict locdict inner_nodes reverse_locdict
	if {![dict exists $typedict $tid]} {
		if {[regexp {^\(([^?]+)\)$} $vs _ s]} {
			append type "\n{$s}"
		} else {
			append type "\n"
		}
		dict set typedict $tid $type
	} else {
		set type [dict get $typedict $tid]
	}
	if {$type == "?\n"} {
		.left.nodes tag configure $nid -background indianred1
	}
	dict set nodedict $nid $tid
	.left.nodes tag bind $nid <Enter> "node_enter $nid"
	.left.nodes tag bind $nid <Leave> "node_leave $nid"
	.left.nodes tag raise $nid
	dict set locdict $nid $loc
	dict lappend reverse_locdict $loc $nid
	foreach n $tag_stack {
		dict lappend inner_nodes $n $nid
	}
	lappend tag_stack $nid
}

proc ftag {id data} {
	global lambdadict
	dict set lambdadict $id $data
}

proc etags {} {
	global tag_stack
	set tag_stack [lrange $tag_stack 0 end-1]
}

proc title {str} {
	.right.box.events.text insert end $str title
	.right.box.events.text see end
}

proc select_file {} {
	ensure_file [.right.files get]
}

proc unhighlight {} {
	global nodedict
	foreach nid [dict keys $nodedict] {
		default_color $nid
	}
}

proc select_loc {x y} {
	global reverse_locdict current_filename nodedict
	unhighlight
	if {![catch	{.right.box.source.code index @$x,$y} ix]} {
		regexp {^(\d+)} $ix _ line
		set key "$current_filename:$line"
		.right.box.source.code tag remove location 1.0 end
		mark_src $key 0
		if {[dict exists $reverse_locdict $key]} {
			set nids [dict get $reverse_locdict $key]
			.left.nodes see [lindex $nids 0].first
			foreach nid $nids {
				select_node $nid
			}
		}
	}
}

proc ensure_file {str} {
	global current_filename pseudo_files
	if {$str != $current_filename} {
		if {[dict exists $pseudo_files $str]} {
			set current_filename $str
			.right.files set $str
			set txt [dict get $pseudo_files $str]
			.right.box.source.code insert end $txt
			return 1
		}
		if {[file exists $str]} {
			set current_filename $str
			.right.files set $str
			set f [open $str]
			set txt [read $f]
			close $f
			.right.box.source.code delete 1.0 end
			.right.box.source.code insert end $txt
			return 1
		}
		return 0
	}
	return 1
}

proc srcfile {name} {
	global all_files current_filename
	lappend all_files $name
	.right.files configure -values $all_files
	ensure_file $name
}

proc src {line} {
	global current_filename pseudo_files
	dict append pseudo_files $current_filename "$line\n"
}

proc mark_lambdas {w withdata} {
	global lambdadict
	set found [$w search -all -regexp -count lengths {<f\d+>} 1.0 end]
	if {$found == ""} return
	foreach f $found n $lengths {
		set id [$w get $f "$f+$n chars"]
		if {[dict exists $lambdadict $id]} {
			$w tag add $id $f "$f+$n chars"
			$w tag raise $id
			$w tag bind $id <Enter> "function_enter $id $withdata"
			$w tag bind $id <Leave> "function_leave $id"
		}
	}
}

proc complete_tags {} {
	mark_lambdas .left.nodes 1
	.left.nodes tag raise sel
	.right.box.source.code tag raise sel
	.right.box.events.text tag raise sel
}

proc function_enter {id withdata} {
	global lambdadict current_filename
	.left.nodes tag configure $id -background lightblue1
	if {$withdata} {
		set current_filename ""
		.right.box.source.code delete 1.0 end
		.right.box.source.code insert 1.0 "procedure info for $id\n"
		.right.box.source.code insert end [dict get $lambdadict $id]
		mark_lambdas .right.box.source.code 0
		mark_lambdas .left.nodes 1
	}
}

proc function_leave {id} {
	global lambdadict
	.left.nodes tag configure $id -background white
}

setup_ui
fconfigure stdin -buffering line -blocking 0
fileevent stdin readable eval_input
vwait forever
